@@ -1,9 +1,11 @@
 import json
 import subprocess
 import functools
+import os
 import os.path
 import shutil
 import sys
+import datetime
 import time
 import re
 import json
@@ -81,6 +83,11 @@ class BandwidthMonitor(object):
     self.prev_time = None
     self.prev_stats = None
     self.bandwidth = None
+    self._prev_received = 0
+    self._prev_sent = 0
+    self._overflow_received = 0
+    self._overflow_sent = 0
+
     self.update()
 
   def current_stats(self):
@@ -113,9 +120,15 @@ class BandwidthMonitor(object):
       m = self.devre.match(line)
       if m and m.group(1) == self.device:
         fields = m.group(2).split()
-        received = fields[0]
-        sent = fields[8]
-        return [int(received), int(sent)]
+        received = long(fields[0])
+        sent = long(fields[8])
+        if self._prev_received > received:
+          self._overflow_received += 2**32
+        self._prev_received = received
+        if self._prev_sent > sent:
+          self._overflow_sent += 2**32
+        self._prev_sent = sent
+        return [received + self._overflow_received, sent + self._overflow_sent ]
     return None
 
 
@@ -127,6 +140,9 @@ class Warrior(object):
     self.warrior_hq_url = warrior_hq_url
     self.real_shutdown = real_shutdown
 
+    # disable the password prompts
+    self.gitenv = dict( os.environ.items() + { 'GIT_ASKPASS': 'echo', 'SSH_ASKPASS': 'echo' }.items() )
+
     self.warrior_id = StringConfigValue(
       name="warrior_id",
       title="Warrior ID",
@@ -164,10 +180,11 @@ class Warrior(object):
     self.bandwidth_monitor = BandwidthMonitor("eth0")
     self.bandwidth_monitor.update()
 
+    self.runner = Runner(concurrent_items=self.concurrent_items)
+    self.runner.on_finish += self.handle_runner_finish
+
     self.current_project_name = None
     self.current_project = None
-    self.current_pipeline = None
-    self.current_runner = None
 
     self.selected_project = None
 
@@ -191,6 +208,7 @@ class Warrior(object):
 
     self.hq_updater = ioloop.PeriodicCallback(self.update_warrior_hq, 10*60*1000)
     self.project_updater = ioloop.PeriodicCallback(self.update_project, 60*60*1000)
+    self.forced_reboot_timeout = None
 
     self.lat_lng = None
     self.find_lat_lng()
@@ -245,6 +263,9 @@ class Warrior(object):
         # time for an update
         print "Reboot for Seesaw update."
         self.reboot_gracefully()
+
+        # schedule a forced reboot after two days
+        self.schedule_forced_reboot()
         return
 
       projects_list = data["projects"]
@@ -292,12 +313,21 @@ class Warrior(object):
         self.failed_projects.discard(project_name)
 
       if os.path.exists(project_path):
+        subprocess.Popen(
+            args=[ "git", "config", "remote.origin.url", project["repository"] ],
+            cwd=project_path
+        ).communicate()
+
         p = AsyncPopen(
             args=[ "git", "pull" ],
-            cwd=project_path
+            cwd=project_path,
+            env=self.gitenv
         )
       else:
-        p = AsyncPopen(args=[ "git", "clone", project["repository"], project_path ])
+        p = AsyncPopen(
+            args=[ "git", "clone", project["repository"], project_path ],
+            env=self.gitenv
+        )
       p.on_output += self.collect_install_output
       p.on_end += yield gen.Callback("gitend")
       p.run()
@@ -356,8 +386,7 @@ class Warrior(object):
   def update_project(self):
     if self.selected_project and (yield gen.Task(self.check_project_has_update, self.selected_project)):
       # restart project
-      if self.current_runner:
-        self.current_runner.stop_gracefully()
+      self.start_selected_project()
 
   @gen.engine
   def check_project_has_update(self, project_name, callback):
@@ -371,9 +400,15 @@ class Warrior(object):
         callback(True)
         return
 
+      subprocess.Popen(
+          args=[ "git", "config", "remote.origin.url", project["repository"] ],
+          cwd=project_path
+      ).communicate()
+
       p = AsyncPopen(
           args=[ "git", "fetch" ],
-          cwd=project_path
+          cwd=project_path,
+          env=self.gitenv
       )
       p.on_output += self.collect_install_output
       p.on_end += yield gen.Callback("gitend")
@@ -414,6 +449,25 @@ class Warrior(object):
       self.on_project_selected(self, project_name)
       self.start_selected_project()
 
+  def clone_project(self, project_name, project_path):
+    version_string = subprocess.Popen(
+        args=[ "git", "log", "-1", "--pretty=%h" ],
+        cwd=project_path,
+        stdout=subprocess.PIPE
+    ).communicate()[0].strip()
+    
+    project_versioned_path = os.path.join(self.data_dir, "projects", "%s-%s" % (project_name, version_string))
+    if not os.path.exists(project_versioned_path):
+      if not os.path.exists(os.path.join(self.data_dir, "projects")):
+        os.makedirs(os.path.join(self.data_dir, "projects"))
+
+      subprocess.Popen(
+          args=[ "git", "clone", project_path, project_versioned_path ],
+          env=self.gitenv
+      ).communicate()
+
+    return project_versioned_path
+
   def load_pipeline(self, pipeline_path, context):
     dirname, basename = os.path.split(pipeline_path)
     if dirname == "":
@@ -441,44 +495,53 @@ class Warrior(object):
   def start_selected_project(self):
     project_name = self.selected_project
 
-    if self.current_project_name == project_name:
-      # already running
-      return
-
-    if self.current_runner:
-      self.current_runner.stop_gracefully()
-      self.fire_status()
-      return
-
     if project_name in self.projects:
+      # install or update project if necessary
       if not project_name in self.installed_projects or (yield gen.Task(self.check_project_has_update, project_name)):
         result = yield gen.Task(self.install_project, project_name)
         if not result:
           return
 
-      project = self.projects[self.selected_project]
+      # remove the configuration variables from the previous project
+      if self.current_project:
+        for config_value in self.current_project.config_values:
+          self.config_manager.remove(config_value.name)
 
+      # the path with the project code
+      # (this is the most recent code from the repository)
       project_path = os.path.join(self.projects_dir, project_name)
-      pipeline_path = os.path.join(project_path, "pipeline.py")
 
+      # clone the project code to a versioned directory
+      # where the pipeline is actually run
+      project_versioned_path = self.clone_project(project_name, project_path)
+
+      # load the pipeline from the versioned directory
+      pipeline_path = os.path.join(project_versioned_path, "pipeline.py")
       (project, pipeline, config_values) = self.load_pipeline(pipeline_path, { "downloader": self.downloader })
 
+      # add the configuration values to the config manager
       for config_value in config_values:
         self.config_manager.add(config_value)
       project.config_values = config_values
 
-      runner = Runner(pipeline, concurrent_items=self.concurrent_items)
-      runner.on_finish += self.handle_runner_finish
+      # start the pipeline
+      if not self.shut_down_flag and not self.reboot_flag:
+        self.runner.set_current_pipeline(pipeline)
 
       self.current_project_name = project_name
       self.current_project = project
-      self.current_pipeline = pipeline
-      self.current_runner = runner
 
-      self.on_project_refresh(self, self.current_project, self.current_runner)
+      self.on_project_refresh(self, self.current_project, self.runner)
       self.fire_status()
 
-      runner.start()
+      if not self.shut_down_flag and not self.reboot_flag:
+        self.runner.start()
+
+    else:
+      # project_name not in self.projects,
+      # stop the current project (if there is one)
+      self.runner.set_current_pipeline(None)
+      self.fire_status()
 
   def handle_runner_finish(self, runner):
     if self.current_project:
@@ -487,10 +550,8 @@ class Warrior(object):
 
     self.current_project_name = None
     self.current_project = None
-    self.current_pipeline = None
-    self.current_runner = None
 
-    self.on_project_refresh(self, self.current_project, self.current_runner)
+    self.on_project_refresh(self, self.current_project, self.runner)
     self.fire_status()
 
     if self.shut_down_flag or self.reboot_flag:
@@ -502,9 +563,6 @@ class Warrior(object):
         elif self.reboot_flag:
           os.system("sudo shutdown -r now")
 
-    elif self.selected_project:
-      self.start_selected_project()
-
   def start(self):
     self.hq_updater.start()
     self.project_updater.start()
@@ -515,19 +573,27 @@ class Warrior(object):
     self.shut_down_flag = False
     self.reboot_flag = True
     self.fire_status()
-    if self.current_runner:
-      self.current_runner.stop_gracefully()
+    if self.runner.is_active():
+      self.runner.set_current_pipeline(None)
     else:
       ioloop.IOLoop.instance().stop()
       if self.real_shutdown:
         os.system("sudo shutdown -r now")
 
+  def schedule_forced_reboot(self):
+    if self.real_shutdown and not self.forced_reboot_timeout:
+      self.forced_reboot_timeout = ioloop.IOLoop.instance().add_timeout(datetime.timedelta(days=2), self.forced_reboot)
+
+  def forced_reboot(self):
+    if self.real_shutdown:
+      os.system("sudo shutdown -r now")
+
   def stop_gracefully(self):
     self.shut_down_flag = True
     self.reboot_flag = False
     self.fire_status()
-    if self.current_runner:
-      self.current_runner.stop_gracefully()
+    if self.runner.is_active():
+      self.runner.set_current_pipeline(None)
     else:
       ioloop.IOLoop.instance().stop()
       if self.real_shutdown:
@@ -536,8 +602,7 @@ class Warrior(object):
   def keep_running(self):
     self.shut_down_flag = False
     self.reboot_flag = False
-    if self.current_runner:
-      self.current_runner.keep_running()
+    self.start_selected_project()
     self.fire_status()
 
   class Status(object):
@@ -565,12 +630,7 @@ class Warrior(object):
       return Warrior.Status.NO_PROJECT
     elif self.selected_project:
       if self.selected_project == self.current_project_name:
-        if self.current_runner.should_stop():
-          return Warrior.Status.RESTARTING_PROJECT
-        else:
-          return Warrior.Status.RUNNING_PROJECT
-      elif self.current_runner:
-        return Warrior.Status.SWITCHING_PROJECT
+        return Warrior.Status.RUNNING_PROJECT
       else:
         return Warrior.Status.STARTING_PROJECT
     else:
